/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.konfiguration;

import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Es wird ein Strassensegment modelliert.
 * 
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class StrassenSegment extends DavProxyObjekt {
	/**
	 * Debug-Logger für Logging-Ausgaben
	 */
	private final static Debug _debug = Debug.getLogger();

	/**
	 * Erstellt eine neue Instanz vom Typ <code>StrassenSegment</code>
	 * 
	 * @param davObjekt
	 *            das zuzuordnente Dav-Objekt
	 * @param laenge
	 *            die Länge in m des Strassensegments
	 */
	protected StrassenSegment(SystemObject davObjekt, long laenge) {
		super(davObjekt);
		_laenge = laenge;
	}

	/**
	 * Länge in m dieses Straßensegments
	 */
	private long _laenge;

	/**
	 * @return die Länge in m dieses Straßensegments
	 */
	public long getLaenge() {
		return _laenge;
	}

	/**
	 * Die Nachfolger-Strassensegmente zu diesem Strassensegment
	 */
	protected Set<StrassenSegment> _nachfolger = new LinkedHashSet<StrassenSegment>();

	/**
	 * Read-Only-Zugriff auf die Nachfolger-Strassensegmente zu diesem
	 * Strassensegment
	 * 
	 * @return Sammlung der Nachfolger-StrassenSegmente
	 */
	public Set<StrassenSegment> getNachFolger() {
		return Collections.unmodifiableSet(_nachfolger);
	}

	/**
	 * Fügt diesem Segment ein Nachfolger-Segment hinzu
	 * 
	 * @param segment
	 *            das Nachfolger-Segment
	 */
	void addNachfolger(StrassenSegment segment) {
		_nachfolger.add(segment);
	}

	/**
	 * Die Vorgänger-Strassensegmente zu diesem Strassensegment
	 */
	protected Set<StrassenSegment> _vorgaenger = new LinkedHashSet<StrassenSegment>();

	/**
	 * Read-Only-Zugriff auf die Vorgänger-Strassensegmente zu diesem
	 * Strassensegment
	 * 
	 * @return Sammlung der Vorgänger-StrassenSegmente
	 */
	public Set<StrassenSegment> getVorgaenger() {
		return Collections.unmodifiableSet(_vorgaenger);
	}

	/**
	 * Fügt diesem Segment ein Vorgänger-Segment hinzu
	 * 
	 * @param segment
	 *            das Vorgänger-Segment
	 */
	void addVorgaenger(StrassenSegment segment) {
		_vorgaenger.add(segment);
	}

	/**
	 * Die Nachfolger-Strassensegmente auf derselben Straße zu diesem
	 * Strassensegment
	 */
	protected Set<StrassenSegment> _nachfolgerAufStrasse = new LinkedHashSet<StrassenSegment>();

	/**
	 * Read-Only-Zugriff auf die Nachfolger-Strassensegmente auf derselben
	 * Straße zu diesem Strassensegment
	 * 
	 * @return Sammlung der Nachfolger-StrassenSegmente auf derselben Straße
	 */
	public Set<StrassenSegment> getNachfolgerAufStrasse() {
		return Collections.unmodifiableSet(_nachfolgerAufStrasse);
	}

	/**
	 * Fügt diesem Segment ein Nachfolger-Segment auf derselben Straße hinzu
	 * 
	 * @param segment
	 *            das Nachfolger-Segment auf derselben Straße
	 */
	void addNachfolgerAufStrasse(StrassenSegment segment) {
		_nachfolgerAufStrasse.add(segment);
	}

	/**
	 * Die Vorgänger-Strassensegmente auf derselben Straße zu diesem
	 * Strassensegment
	 */
	protected Set<StrassenSegment> _vorgaengerAufStrasse = new LinkedHashSet<StrassenSegment>();

	/**
	 * Read-Only-Zugriff auf die Vorgänger-Strassensegmente auf derselben Straße
	 * zu diesem Strassensegment
	 * 
	 * @return Sammlung der Vorgänger-StrassenSegmente auf derselben Straße
	 */
	public Set<StrassenSegment> getVorgaengerAufStrasse() {
		return Collections.unmodifiableSet(_vorgaengerAufStrasse);
	}

	/**
	 * Fügt diesem Segment ein Vorgänger-Segment auf derselben Straße hinzu
	 * 
	 * @param segment
	 *            das Vorgänger-Segment auf derselben Straße
	 */
	void addVorgaengerAufStrasse(StrassenSegment segment) {
		_vorgaengerAufStrasse.add(segment);
	}

	/**
	 * @return das auf der Strasse vorhergehende Strassensegment mit der
	 *         kleinsten Länge oder null, falls kein Vorgänger auf Strasse
	 *         existiert - wennn es mehrere Segmente mit derselben Länge gibt,
	 *         wird das erste zurückgegeben
	 */
	public StrassenSegment getKuerzestenVorgaengerAufStrasse() {
		StrassenSegment result = null;
		for (StrassenSegment segment : getVorgaengerAufStrasse()) {
			if (result == null) {
				result = segment;
			} else {
				if (segment.getLaenge() < result.getLaenge()) {
					result = segment;
				}
			}
		}
		return result;
	}

	/**
	 * @return das auf der Strasse nachfolgende Strassensegment mit der
	 *         kleinsten Länge oder null, falls kein Nachfolger auf Strasse
	 *         existiert - wennn es mehrere Segmente mit derselben Länge gibt,
	 *         wird das erste zurückgegeben
	 */
	public StrassenSegment getKuerzestenNachfolgerAufStrasse() {
		StrassenSegment result = null;
		for (StrassenSegment segment : getNachfolgerAufStrasse()) {
			if (result == null) {
				result = segment;
			} else {
				if (segment.getLaenge() < result.getLaenge()) {
					result = segment;
				}
			}
		}
		return result;
	}

	/**
	 * Methode prüft bei äußeren Straßensegmenten ob das Straßensegment in einem
	 * Autobahnkreuz oder Autobahndreieck beginnt. Bei inneren Straßensegmenten
	 * wird geprüft ob der Straßenknoten in dem das Straßensegment liegt ein
	 * Autobahnkreuz oder Autobahndreieck ist. (Methode wird in den abgeleiteten
	 * Klassen überschrieben)
	 * 
	 * @return Autobahnkreuz, Autobahndreick: true, sonst false
	 */
	public boolean beginntInAutobahnKreuzOderDreieck() {
		return false;
	}

	/**
	 * Methode prüft bei äußeren Straßensegmenten ob das Straßensegment in einem
	 * Autobahnkreuz oder Autobahndreieck endet. Bei inneren Straßensegmenten
	 * wird geprüft ob der Straßenknoten in dem das Straßensegment liegt ein
	 * Autobahnkreuz oder Autobahndreieck ist. (Methode wird in den abgeleiteten
	 * Klassen überschrieben)
	 * 
	 * @return Autobahnkreuz, Autobahndreick: true, sonst false
	 */
	public boolean endetInAutobahnKreuzOderDreieck() {
		return false;
	}

	/**
	 * Methode prüft bei äußeren Straßensegmenten ob das Segment am
	 * Autobahnanfang beginnt. Bei inneren Straßensegmenten wird geprüft, ob der
	 * Straßenknoten in dem das Straßensegment liegt ein Autobahnanfang ist.
	 * (Methode wird in den abgeleiteten Klassen überschrieben)
	 * 
	 * @return Autobahnanfang: true, sonst false
	 */
	public boolean beginntAmAutobahnAnfang() {
		return false;
	}

	/**
	 * Methode prüft bei äußeren Straßensegmenten ob das Segment am Autobahnende
	 * endet. Bei inneren Straßensegmenten wird geprüft, ob der Straßenknoten in
	 * dem das Straßensegment liegt ein Autobahnende ist. (Methode wird in den
	 * abgeleiteten Klassen überschrieben)
	 * 
	 * @return Autobahnende: true, sonst false
	 */
	public boolean endetAmAutobahnEnde() {
		return false;
	}

	/**
	 * Comparator, der eine Reihenfolge von <code>MessQuerschnitt</code>-
	 * Instanzen nach dem Wert des Attributs Offset vorgibt
	 * 
	 * @author Dambach Werke GmbH
	 */
	static class MqComparator implements Comparator<MessQuerschnitt> {
		/*
		 * (Kein Javadoc)
		 * 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(MessQuerschnitt mq1, MessQuerschnitt mq2) {
			if (mq1.getOffset() < mq2.getOffset()) {
				return -1;
			} else if (mq1.getOffset() > mq2.getOffset()) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * Die Instanz von <code>MqComparator</code>, die für alle Strassensegmente
	 * benutzt wird
	 */
	static private MqComparator _mqComparator = new MqComparator();

	/**
	 * Die Messquerschnitte, die auf diesem Strassensegment liegen
	 */
	SortedSet<MessQuerschnitt> _messQuerschnitte = new TreeSet<MessQuerschnitt>(_mqComparator);

	/**
	 * Read only Zugriff auf die Messquerschnitte, die auf diesem
	 * Strassensegment liegen
	 * 
	 * @return die Messquerschnitte, die auf diesem Strassensegment liegen -
	 *         sortiert aufsteigend nach Offset
	 */
	public SortedSet<MessQuerschnitt> getMessQuerschnitte() {
		return Collections.unmodifiableSortedSet(_messQuerschnitte);
	}

	/**
	 * Fügt den Messquerschnitten auf diesem Strassensegment einen hinzu
	 * 
	 * @param mq
	 *            der hinzuzufügende Messquerschnitt
	 */
	void addMessQuerschnitt(MessQuerschnitt mq) {
		_messQuerschnitte.add(mq);
	}

	/**
	 * Die Strassenteilsegmente dieses Strassensegments
	 */
	List<StrassenTeilSegment> _strassenTeilSegmente = new LinkedList<StrassenTeilSegment>();

	/**
	 * Read only Zugriff auf die Strassenteilsegmente dieses Strassensegments
	 * 
	 * @return die Strassenteilsegmente dieses Strassensegments
	 */
	public List<StrassenTeilSegment> getStrassenTeilSegmente() {
		return Collections.unmodifiableList(_strassenTeilSegmente);
	}

	/**
	 * Fügt den Strassenteilsegmenten dieses Strassensegments eines hinzu
	 * 
	 * @param strassenTeilSegment
	 *            das neue Strassenteilsegment
	 */
	void addStrassenTeilSegment(StrassenTeilSegment strassenTeilSegment) {
		_strassenTeilSegmente.add(strassenTeilSegment);
	}

}
