/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.meldungserstellung;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.tmc.allgemein.RDSEreignis;
import de.bsvrz.ste.tmc.konfiguration.TmcLocationPunkt;

/**
 * Ereignis.
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class Ereignis {
	public enum Art {
		NULL, NORMAL, INITIAL;
	}

	private Art _art;
	private RDSEreignis _rdsEreignis;
	private List<TmcLocationPunkt> _tmcOrte = new ArrayList<TmcLocationPunkt>();
	private int _tmcRichtung;
	private String _ereignisCode;
	private RDSQuantitaetLEN _rdsQuantitaetLEN = null;
	private SystemObject _erklaerungsKomponente = null;
	private List<EreignisDaten> _datenListe = null;

	Ereignis(Art art, RDSEreignis rdsEreignis, Ereignis ereignis) {
		this(art, rdsEreignis, ereignis.getTmcOrte(), ereignis.getTmcRichtung(), ereignis.getEreignisCode());
		setRDSQuantitaetLEN(ereignis.getRDSQuantitaetLEN());
		setErklaerungsKomponente(ereignis.getErklaerungsKomponente());
		setDatenListe(ereignis.getDatenListe());
	}

	public Ereignis(Art art, RDSEreignis rdsEreignis, List<TmcLocationPunkt> tmcOrte, int tmcRichtung,
			String ereignisCode) {
		_art = art;
		_rdsEreignis = rdsEreignis;
		_tmcOrte.addAll(tmcOrte);
		_tmcRichtung = tmcRichtung > 0 ? 1 : 0;
		_ereignisCode = ereignisCode;
	}

	@Override
	public boolean equals(final Object objekt) {
		if (!(objekt instanceof Ereignis)) {
			return false;
		}
		final Ereignis ereignis = (Ereignis) objekt;

		// Ereignisse unterschiedlicher Art sind grundsätzlich unterschiedlich
		// Ausnahme: zwei Ereignisse mit je der Art NORMAL und INITIAL können
		// gleich sein
		boolean result = ((getArt().compareTo(ereignis.getArt()) == 0)
				|| (getArt().compareTo(Art.NORMAL) == 0 && ereignis.getArt().compareTo(Art.INITIAL) == 0)
				|| (getArt().compareTo(Art.INITIAL) == 0 && ereignis.getArt().compareTo(Art.NORMAL) == 0));

		if (!result) {
			return result;
		}

		result = getRDSEreignis().compareTo(ereignis.getRDSEreignis()) == 0
				&& getTmcRichtung() == ereignis.getTmcRichtung() && getTmcOrte().equals(ereignis.getTmcOrte())
				&& ((getEreignisCode() != null && getEreignisCode().equals(ereignis.getEreignisCode()))
						|| (getEreignisCode() == null && ereignis.getEreignisCode() == null))
				&& ((getRDSQuantitaetLEN() != null && getRDSQuantitaetLEN().equals(ereignis.getRDSQuantitaetLEN()))
						|| (getRDSQuantitaetLEN() == null && ereignis.getRDSQuantitaetLEN() == null));

		if (!result || (getArt().compareTo(Art.NORMAL) == 0 && ereignis.getArt().compareTo(Art.INITIAL) == 0)
				|| (getArt().compareTo(Art.INITIAL) == 0 && ereignis.getArt().compareTo(Art.NORMAL) == 0)) {
			// Wenn zwei Ereignisse mit je der Art NORMAL und INITIAL (oder
			// umgekehrt) verglichen werden,
			// hört der Vergleich hier auf
			return result;
		}

		result = ((getErklaerungsKomponente() != null
				&& getErklaerungsKomponente().equals(ereignis.getErklaerungsKomponente()))
				|| (getErklaerungsKomponente() == null && ereignis.getErklaerungsKomponente() == null))
				&& ((getDatenBeschreibung() != null && getDatenBeschreibung().equals(ereignis.getDatenBeschreibung()))
						|| (getDatenBeschreibung() == null && ereignis.getDatenBeschreibung() == null));

		return result;
	}

	public String getDatenBeschreibung() {
		String result = null;
		if (getDatenListe() != null) {
			result = "TMC Ereignis '" + getRDSEreignis() + "':";
			for (EreignisDaten daten : getDatenListe()) {
				result += "\n\tLänge " + daten.getLaenge() + "m beginnend bei " + daten.start + "m auf Segment "
						+ daten.startSegment.getDavObjekt().getPid();
				if (daten.lokal != null) {
					result += " (" + daten.lokal.getDavObjekt().getPid() + ")";
				}
			}
		}

		return result;
	}

	@Override
	public String toString() {
		String result = "\"";
		result += getEreignisCode() != null ? getEreignisText() + "\" (" + getEreignisCode() + ")" : getRDSEreignis();
		result += getTmcRichtung() > 0 ? " in negativer Richtung" : " in positiver Richtung";
		result += " für TMC-Knoten " + getTmcPrimaererOrt().getLocationCode();
		for (int i = 1; i < getTmcOrte().size(); i += 1) {
			result += ", " + getTmcOrte().get(i).getLocationCode();
		}
		if (getRDSQuantitaetLEN() != null) {
			result += " - Ereignislänge Wert: " + getRDSQuantitaetLEN().getBetroffeneLaenge() + "["
					+ RDSQuantitaetLEN.EINHEIT + "]";
		}
		result += "\"";
		return result;
	}

	public Art getArt() {
		return _art;
	}

	public RDSEreignis getRDSEreignis() {
		return _rdsEreignis;
	}

	public List<TmcLocationPunkt> getTmcOrte() {
		return Collections.unmodifiableList(_tmcOrte);
	}

	public TmcLocationPunkt getTmcPrimaererOrt() {
		return !_tmcOrte.isEmpty() ? _tmcOrte.get(0) : null;
	}

	/**
	 * @return Richtung der primären TMC-Location - Ereignisrichtung ist stets
	 *         entgegengesetzt
	 */
	public int getTmcRichtung() {
		return _tmcRichtung;
	}

	public int getAusdehnung() {
		return _tmcOrte.size() - 1;
	}

	public String getEreignisCode() {
		return _ereignisCode;
	}

	public String getEreignisText() {
		return RDSBeschreibung.getTextZuCode(_rdsEreignis, _ereignisCode, _erklaerungsKomponente != null);
	}

	public String getSchluessel() {
		return getTmcOrte().get(0).getLocationCode() + (getTmcRichtung() < 0 ? "-" : getTmcRichtung() > 0 ? "+" : "");
	}

	public RDSQuantitaetLEN getRDSQuantitaetLEN() {
		return _rdsQuantitaetLEN;
	}

	public void setRDSQuantitaetLEN(RDSQuantitaetLEN rdsQuantitaetLEN) {
		_rdsQuantitaetLEN = rdsQuantitaetLEN;
	}

	public SystemObject getErklaerungsKomponente() {
		return _erklaerungsKomponente;
	}

	public void setErklaerungsKomponente(SystemObject erklaerungsKomponente) {
		_erklaerungsKomponente = erklaerungsKomponente;
	}

	protected List<EreignisDaten> getDatenListe() {
		return _datenListe;
	}

	protected void setDatenListe(List<EreignisDaten> datenListe) {
		_datenListe = datenListe;
	}

	public String getAlertCCode() {
		// 1. Byte Multi Group Message
		// String code = "00";

		// 1. Byte Single Group Message
		String code = "08";

		int diversion = 0x01 & 0; // 1 Bit
		int direction = 0x01 & getTmcRichtung(); // 1 Bit
		int extend = 0x07 & getAusdehnung(); // 3 Bit
		int event;
		try {
			event = Integer.parseInt(getEreignisCode());
		} catch (NumberFormatException e) {
			event = 0;
		}

		// 2. Byte 1 Bit Umleitungsempfehlung, 1 Bit Richtung, 3 Bit Extend, die
		// ersten 3 Bit Ereigniscode
		code += String.format("%02X", (diversion << 7) | (direction << 6) | (extend << 3) | ((0x0700 & event) >> 8));

		// 3. Byte Ereigniscode restliche 8 Bit
		code += String.format("%02X", 0xFF & event);

		int location = getTmcPrimaererOrt() != null ? getTmcPrimaererOrt().getLocationCode() : 0;
		// 4. und 5 Byte: Location 16 Bit
		code += String.format("%02X", location >> 8);
		code += String.format("%02X", 0xFF & location);

		for (int i = code.length() / 2; i < 25; i += 1) {
			code += "00";
		}

		return code;
	}

}
