/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.meldungserstellung;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ste.tmc.allgemein.RDSEreignis;
import de.bsvrz.ste.tmc.konfiguration.StrassenSegment;
import de.bsvrz.ste.tmc.parameter.RDSParameterDaten;
import de.bsvrz.ste.tmc.verkehrsanalyse.StoerfallSituation;
import de.bsvrz.ste.tmc.verkehrsanalyse.StoerfallSituationDaten;
import de.bsvrz.ste.tmc.verkehrsanalyse.StoerfallSituationListener;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans.
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class Route implements StoerfallSituationListener

{
	/**
	 * Debug-Logger für Logging-Ausgaben
	 */
	private final static Debug _debug = Debug.getLogger();

	/**
	 * Comparator, der eine Reihenfolge von <code>EreignisDaten</code> nach den
	 * Werten der Attribute offset und start vorgibt
	 * 
	 * @author Dambach Werke GmbH
	 */
	static class EreignisDatenComparator implements Comparator<EreignisDaten> {
		/*
		 * (Kein Javadoc)
		 * 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(EreignisDaten d1, EreignisDaten d2) {
			if (d1.startSegmentOffset + d1.start < d2.startSegmentOffset + d2.start) {
				return -1;
			} else if (d1.startSegmentOffset + d1.start > d2.startSegmentOffset + d2.start) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	/**
	 * Instanz von <code>EreignisDatenComparator</code>, die für alle
	 * Störfallsituationen benutzt wird
	 */
	static private EreignisDatenComparator EREIGNIS_DATEN_COMPARATOR = new EreignisDatenComparator();

	private List<StrassenSegment> _segmente = new LinkedList<StrassenSegment>();

	public List<StrassenSegment> getSegmente() {
		return Collections.unmodifiableList(_segmente);
	}

	Meldungserstellung _meldungserstellung;

	public RDSParameterDaten getParameter() {
		return _meldungserstellung.getParameter();
	}

	public Route(Meldungserstellung meldungserstellung, List<StrassenSegment> segmente) {
		_meldungserstellung = meldungserstellung;
		_segmente.addAll(segmente);
	}

	private Map<StrassenSegment, List<StoerfallSituationDaten>> _stoerfaelleAufSegmenten = new LinkedHashMap<StrassenSegment, List<StoerfallSituationDaten>>();

	List<StoerfallSituationDaten> getStoerfalleAufSegment(StrassenSegment segment) {
		return _stoerfaelleAufSegmenten.get(segment);
	}

	public void neueStoerfallSituation(final StoerfallSituation quelle, final List<StoerfallSituationDaten> daten) {
		if (getSegmente().contains(quelle.getSegment())) {
			synchronized (_meldungserstellung.mutexEreignis) {
				_stoerfaelleAufSegmenten.put(quelle.getSegment(), daten);
				fireEreignisListener();
			}
		} else {
			throw new IllegalArgumentException("Straßensegment " + quelle.getSegment()
					+ " wurde an eine Route übergeben, die dieses Segment nicht enthält");
		}
	}

	// Bestimme Ereignisdaten für die ganze Route
	private List<EreignisDaten> abbildung() {
		final List<EreignisDaten> result = new ArrayList<EreignisDaten>();

		long offset = 0;

		for (StrassenSegment segment : getSegmente()) {
			final List<StoerfallSituationDaten> liste = _stoerfaelleAufSegmenten.get(segment);

			if (liste != null) {
				for (StoerfallSituationDaten daten : liste) {
					final RDSEreignis ereignis = getParameter().getRDSAbbildungEreignis().get(daten.zustand);
					if (ereignis != null) {
						result.add(new EreignisDaten(offset, offset, segment, segment, ereignis, daten.start,
								daten.ende, daten.lokal));
					}
				}
			}

			offset += segment.getLaenge();
		}

		// for (StrassenSegment segment : _stoerfaelleAufSegmenten.keySet()) {
		// // Bestimme Offset
		// long offset = 0;
		// for (StrassenSegment lauf : getSegmente()) {
		// if (lauf == segment) {
		// break;
		// }
		//
		// }
		//
		// // Bestimme Ereignisdaten für die ganze Route
		// List<StoerfallSituationDaten> liste =
		// _stoerfaelleAufSegmenten.get(segment);
		//
		// for (StoerfallSituationDaten daten : liste) {
		// RDSEreignis ereignis =
		// getParameter().getRDSAbbildungEreignis().get(daten.zustand);
		// if (ereignis != null) {
		// result.add(new EreignisDaten(offset, offset, segment, segment,
		// ereignis, daten.start, daten.ende,
		// daten.lokal));
		// }
		// }
		// }
		// Collections.sort(result, EREIGNIS_DATEN_COMPARATOR);

		return result;
	}

	long getOffset(StrassenSegment segment) {
		long result = 0;
		for (StrassenSegment routenSegment : getSegmente()) {
			if (routenSegment == segment) {
				return result;
			} else {
				result += routenSegment.getLaenge();
			}
		}
		throw new IllegalArgumentException("Segment " + segment.getDavObjekt().getPid() + " in Route nicht vorhanden");
	}

	/**
	 * Sammlung aller <code>EreignisDatenListener</code>
	 */
	private Collection<EreignisListener> _ereignisListener = new ArrayList<EreignisListener>();

	public void addEreignisListener(EreignisListener listener) {
		synchronized (_ereignisListener) {
			_ereignisListener.add(listener);
		}
	}

	public void renmoveEreignisListener(EreignisListener listener) {
		synchronized (_ereignisListener) {
			_ereignisListener.remove(listener);
		}
	}

	private List<EreignisDaten> _ereignisDaten = new LinkedList<EreignisDaten>();

	List<EreignisDaten> getEreignisDaten() {
		return Collections.unmodifiableList(_ereignisDaten);
	}

	/**
	 * Benachrichtige alle angemeldeten Listener vom Typ
	 * <code>EreignisDatenDaten</code>
	 */
	public void fireEreignisListener() {
		List<EreignisDaten> daten = abbildung();
		daten = laengenAbgleich(daten);
		daten = laengenBestimmung(daten);
		/* Stau-Plausibilität: ALT Anfang */
		plausibilisiereDaten(daten);
		/* Stau-Plausibilität: ALT Ende */
		daten = findeSchaerfsteZustaende(daten);

		_ereignisDaten.clear();
		_ereignisDaten.addAll(daten);

		synchronized (_ereignisListener) {
			for (EreignisListener listener : _ereignisListener) {
				listener.neueEreignisse(this, daten);
			}
		}
	}

	private List<EreignisDaten> findeSchaerfsteZustaende(final List<EreignisDaten> datenListe) {
		final List<EreignisDaten> result = new LinkedList<EreignisDaten>();

		final Map<StrassenSegment, RDSEreignis> schaerfsteZustaende = new LinkedHashMap<StrassenSegment, RDSEreignis>();

		_debug.fine("findeSchaerfsteZustaende: " + datenListe);

		for (EreignisDaten daten : datenListe) {
			StrassenSegment segment = daten.startSegment;
			while (segment != null) {
				final RDSEreignis zustand = schaerfsteZustaende.get(segment);

				if (zustand == null || zustand.compareTo(daten.ereignis) < 0) {
					schaerfsteZustaende.put(segment, daten.ereignis);
				}
				if (segment != daten.endSegment) {
					segment = segment.getKuerzestenNachfolgerAufStrasse();
				} else {
					segment = null;
				}
			}
		}

		// schaerfsteZustaende enthält jetzt für jedes vorkommende
		// Strassensegment den schärfsten Zustand

		for (EreignisDaten daten : datenListe) {
			final LinkedList<StrassenSegment> segmentListe = new LinkedList<StrassenSegment>();
			StrassenSegment segment = daten.startSegment;

			while (segment != null) {
				final RDSEreignis zustand = schaerfsteZustaende.get(segment);

				if (daten.ereignis.compareTo(zustand) >= 0) {
					segmentListe.add(segment);
				}
				if (segment != daten.endSegment) {
					segment = segment.getKuerzestenNachfolgerAufStrasse();
				} else {
					segment = null;
				}
			}

			// segmentListe enthält jetzt für das aktuelle Datum die
			// Straßensegmente, deren schärfster Zustand zum Datum passt - oder
			// ist leer, leere Liste bedeutet das entsprechende Element wird
			// vollständig von einem anderen Element mit schärferem Zustand
			// dominiert und dementsprechend eliminiert.
			if (!segmentListe.isEmpty()) {
				final StrassenSegment startSegment = segmentListe.getFirst();
				final StrassenSegment endSegment = segmentListe.getLast();
				if (startSegment == daten.startSegment && endSegment == daten.endSegment) {
					// Es hat sich für dieses Element nichts geändert
					// Element wird unverändert übernommen
					result.add(daten);
				} else {
					// Für dieses Element wurden Segmente weggenommen
					// Es muss ein entsprechendes neues Element erzeugt werden
					final long startSegmentOffset;
					final long start;
					final long endSegmentOffset;
					final long ende;
					if (startSegment == daten.startSegment) {
						// es wurden hinten Segmente weggenommen
						startSegmentOffset = daten.startSegmentOffset;
						start = daten.start;
						endSegmentOffset = getOffset(endSegment);
						ende = endSegment.getLaenge();
					} else if (endSegment == daten.endSegment) {
						// es wurden vorne Segmente weggenommen
						startSegmentOffset = getOffset(startSegment);
						start = 0;
						endSegmentOffset = daten.endSegmentOffset;
						ende = daten.ende;
					} else {
						// es wurden hinten und vorne Segmente weggenommen
						startSegmentOffset = getOffset(startSegment);
						start = 0;
						endSegmentOffset = getOffset(endSegment);
						ende = endSegment.getLaenge();
					}
					result.add(new EreignisDaten(startSegmentOffset, endSegmentOffset, startSegment, endSegment,
							daten.ereignis, start, ende, daten.lokal));
				}
			}
		}

		String text = "ERGEBNIS SCHÄRFSTE ZUSTÄNDE";
		for (EreignisDaten dat : result) {
			text += "\n" + dat.ereignis + ": Länge " + dat.getLaenge() + "m beginnend auf "
					+ dat.startSegment.getDavObjekt().getPid() + " bei " + dat.start + "m";
		}
		_debug.fine(text);

		return result;
	}

	private List<EreignisDaten> laengenAbgleich(List<EreignisDaten> ereignisDaten) {
		LinkedList<EreignisDaten> result = new LinkedList<EreignisDaten>();

		Map<RDSEreignis, List<EreignisDaten>> zustandsMap = new LinkedHashMap<RDSEreignis, List<EreignisDaten>>();
		Map<RDSEreignis, LinkedList<EreignisDaten>> resultMap = new LinkedHashMap<RDSEreignis, LinkedList<EreignisDaten>>();

		// Verteile die EreignisDaten auf RDS-Ereignis-bezogenen Listen und
		// erzeuge ebensolche Ergebnislisten
		for (EreignisDaten daten : ereignisDaten) {
			_debug.finer("laengenAbgleich(1): " + daten);

			final RDSEreignis zustand = daten.ereignis;
			List<EreignisDaten> liste = zustandsMap.get(zustand);
			if (liste == null) {
				liste = new LinkedList<EreignisDaten>();
				zustandsMap.put(zustand, liste);
				resultMap.put(zustand, new LinkedList<EreignisDaten>());
			}
			liste.add(daten);
		}

		for (RDSEreignis zustand : zustandsMap.keySet()) {
			final List<EreignisDaten> liste = zustandsMap.get(zustand);
			final LinkedList<EreignisDaten> ergebnis = resultMap.get(zustand);

			_debug.finer("laengenAbgleich(2.0): " + ergebnis);

			for (EreignisDaten daten : liste) {
				if (ergebnis.isEmpty()) {
					_debug.finer("laengenAbgleich(2.1): " + daten);
					ergebnis.add(new EreignisDaten(daten));
				} else {
					_debug.finer("laengenAbgleich(2.2): " + daten);

					final EreignisDaten vorgaenger = ergebnis.getLast();
					_debug.finer("laengenAbgleich(2.2) Vorgaenger: " + vorgaenger);

					final long start = daten.startSegmentOffset + daten.start;
					final long ende = vorgaenger.endSegmentOffset + vorgaenger.ende;

					_debug.finer("laengenAbgleich(2.3): start: " + start + ", ende: " + ende + ", max: "
							+ getParameter().getRDSDMax());

					if (start - ende < getParameter().getRDSDMax()) {
						final EreignisDaten d = new EreignisDaten(vorgaenger.startSegmentOffset, daten.endSegmentOffset,
								vorgaenger.startSegment, daten.endSegment, vorgaenger.ereignis, vorgaenger.start,
								daten.ende, vorgaenger.lokal == daten.lokal ? vorgaenger.lokal : null);

						_debug.finer("laengenAbgleich(2.4) neue Ereignisdaten: " + d);

						ergebnis.removeLast();
						ergebnis.add(d);
					} else {
						ergebnis.add(new EreignisDaten(daten));
					}
				}
			}
		}

		for (List<EreignisDaten> liste : resultMap.values()) {
			for (EreignisDaten daten : liste) {
				result.add(daten);
			}
		}

		Collections.sort(result, EREIGNIS_DATEN_COMPARATOR);

		_debug.finer("Ergebnis laengenAbgleich: " + result);

		return result;
	}

	private List<EreignisDaten> laengenBestimmung(final List<EreignisDaten> daten) {
		final List<EreignisDaten> result = new LinkedList<EreignisDaten>();
		final int dStoerfallAnfang = getParameter().getRDSDStoerfallAnfang();
		final int dStoerfallEnde = getParameter().getRDSDStoerfallEnde();
		final int mindestLaenge = getParameter().getRDSMindestLaenge();

		for (EreignisDaten lauf : daten) {
			/*
			 * RDS-Ereignis hat Richtung entgegen der Fahrtrichtung - deshalb
			 * wird vor mit dStoerFallEnde (Parameter RDSStoerfallEnde) und nach
			 * mit dStoerFallAnfang (Parameter RDSStoerfallAnfang) berechnet
			 */
			final long vor = dStoerfallEnde > lauf.start ? lauf.start : dStoerfallEnde;
			final long nach = dStoerfallAnfang > lauf.endSegment.getLaenge() - lauf.ende
					? lauf.endSegment.getLaenge() - lauf.ende : dStoerfallAnfang;

			_debug.finer("LaengenBestimmung: " + lauf);
			_debug.finer("LaengenBestimmung: vor: " + vor + ", nach: " + nach, ", endSegmentOffset: "
					+ lauf.endSegmentOffset + ", startSegmentOffset: " + lauf.startSegmentOffset);

			final long laenge = lauf.endSegmentOffset + lauf.ende + nach - (lauf.startSegmentOffset + lauf.start - vor);

			if (laenge >= mindestLaenge) {
				final EreignisDaten ereignisDaten = new EreignisDaten(lauf.startSegmentOffset, lauf.endSegmentOffset,
						lauf.startSegment, lauf.endSegment, lauf.ereignis, lauf.start - vor, lauf.ende + nach,
						lauf.lokal);
				_debug.finer("LaengenBestimmung: neue EreignisDaten:  " + ereignisDaten);
				result.add(ereignisDaten);
			} else {
				_debug.finer(
						"LaengenBestimmung: mindestlaenge " + mindestLaenge + " nicht erreicht mit Laenge: " + laenge);
			}
		}

		return result;
	}

	/* Stau-Plausibilität: ALT Anfang */
	private void plausibilisiereDaten(List<EreignisDaten> daten) {
		if (getParameter().getRDSPlausibStauverlaufsanalyse()) {
			Iterator<EreignisDaten> it = daten.listIterator();

			while (it.hasNext()) {
				EreignisDaten lauf = it.next();
				if (lauf.ereignis.compareTo(RDSEreignis.STAU) == 0) {
					int startIndex = getSegmente().indexOf(lauf.startSegment);
					int endIndex = getSegmente().indexOf(lauf.endSegment);
					if (!_meldungserstellung.getVerkehrsanalyse().existiertStau(lauf.startSegment, lauf.start,
							lauf.endSegment, lauf.ende)) {
						it.remove();
					}
				}
			}
		}
	}
	/* Stau-Plausibilität: ALT Ende */

	SystemObject sucheBaustelle(StrassenSegment wurzelSegment, long offset) {
		if (getParameter().getRDSBeruecksichtigeBaustellen()) {
			long bereichNach = getParameter().getRDSDBaustelleSuchbereichNach();

			long bereich = wurzelSegment.getLaenge() - offset;
			StrassenSegment segment = wurzelSegment;
			StrassenSegment endSegment = segment;
			long endOffset;

			while (segment != null && bereich < bereichNach) {
				segment = segment.getKuerzestenNachfolgerAufStrasse();
				if (segment != null) {
					bereich += segment.getLaenge();
				} else {
					endSegment = segment;
				}
			}

		}

		return null;
	}

	// private SystemObject findeBaustelleNachEreignis(StrassenSegment
	// startSegment, long startOffset,
	// StrassenSegment endSegment, long endOffset)
	// {
	// SystemObject result = null;
	// StrassenSegment segment = startSegment;
	// while (segment != null)
	// {
	// for (Situation baustelle:
	// _meldungserstellung.getBaustellenSegmente().keySet())
	// {
	// SituationEigenschaften eigenschaften =
	// _meldungserstellung.getBaustellenSegmente().get(baustelle);
	// for (int i = 0; result == null && i <
	// eigenschaften.strassenSegment.length; i += 1)
	// {
	// if (eigenschaften.strassenSegment[i].equals(segment.getDavObjekt()))
	// {
	// // Gemeinsames Segment gefunden
	// if (i == 1 && segment == endSegment) // Erstes Baustellensegment ist
	// letztes Suchsegment
	// {
	// if (eigenschaften.startOffset <= endOffset)
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// else if (i == eigenschaften.strassenSegment.length - 1 && segment ==
	// startSegment) // letztes Baustellensegment ist erstes Suchsegment
	// {
	// if (eigenschaften.endOffset >= startOffset)
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// else
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// }
	// }
	//
	// segment = segment != endSegment ?
	// segment.getKuerzestenNachfolgerAufStrasse() : null;
	// }
	// return result;
	// }
	//
	// private SystemObject findeBaustelleVorEreignis(StrassenSegment
	// startSegment, long startOffset,
	// StrassenSegment endSegment, long endOffset)
	// {
	// SystemObject result = null;
	// StrassenSegment segment = endSegment;
	// while (segment != null)
	// {
	// for (Situation baustelle:
	// _meldungserstellung.getBaustellenSegmente().keySet())
	// {
	// SituationEigenschaften eigenschaften =
	// _meldungserstellung.getBaustellenSegmente().get(baustelle);
	// for (int i = eigenschaften.strassenSegment.length - 1; i >= 0; i -= 1)
	// {
	// if (eigenschaften.strassenSegment[i].equals(segment.getDavObjekt()))
	// {
	// // Gemeinsames Segment gefunden
	// if (i == eigenschaften.strassenSegment.length - 1 && segment ==
	// startSegment) // letztes Baustellensegment ist erstes Suchsegment
	// {
	// if (eigenschaften.endOffset >= startOffset)
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// else if (i == 1 && segment == endSegment) // Erstes Baustellensegment ist
	// letztes Suchsegment
	// {
	// if (eigenschaften.startOffset <= endOffset)
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// else
	// {
	// result = eigenschaften.strassenSegment[i];
	// }
	// }
	// }
	// }
	//
	// segment = segment != startSegment ?
	// segment.getKuerzestenVorgaengerAufStrasse() : null;
	// }
	// return result;
	// }

}
