/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ste.tmc.verkehrsanalyse;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.ste.tmc.allgemein.StoerfallZustand;
import de.bsvrz.ste.tmc.konfiguration.StrassenSegment;
import de.bsvrz.ste.tmc.parameter.RDSParameterDaten;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Modelliert die überlagerten Zustände aller Störfallindikatoren eines
 * Strassensegments.
 * 
 * 
 * Erstellt auf Basis der SWE RDS/TMC von: Dambach Werke GmbH, Stefan Sans
 * 
 * @since 07/2016: Adaptiert für TMC-Meldungen durch: BitCtrl Systems GmbH
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class StoerfallSituation implements StoerfallListener {
	/**
	 * Debug-Logger für Logging-Ausgaben.
	 */
	private static final Debug _debug = Debug.getLogger();

	/**
	 * Zeit in ms, nach der eine Überlagerungsberechnung erzwungen wird, falls
	 * die Ereignisqueue nicht leer wird.
	 */
	public final int TIMEOUT = 30000;

	/**
	 * Das äussere Strassensegment dieser Störfallsituation.
	 */
	private StrassenSegment _segment;

	/**
	 * @return das (äussere) Strassensegment dieser Störfallsituation
	 */
	public StrassenSegment getSegment() {
		return _segment;
	}

	/**
	 * Alle Störfalle (Störfallindikatoren) dieser Störfallsituation.
	 */
	private Collection<Stoerfall> _stoerfaelle = new ArrayList<Stoerfall>();

	/**
	 * Sammlung aller <code>StoerfallSituationListener</code>.
	 */
	private Collection<StoerfallSituationListener> _stoerfallSituationListener = new ArrayList<StoerfallSituationListener>();

	/**
	 * Die <code>StoerfallDaten</code> für jeden Störfall mit der höchsten
	 * Priorität je Verfahren.
	 */
	private Map<Stoerfall, StoerfallDaten> _stoerfallDaten = new LinkedHashMap<Stoerfall, StoerfallDaten>();

	/**
	 * Die überlagerten Störfalldaten im zugeordneten Segment.
	 */
	private List<StoerfallDaten> _ueberlagerteDaten = new LinkedList<StoerfallDaten>();

	private Verkehrsanalyse _verkehrsAnalyse;

	Verkehrsanalyse getVerkehrsanalyse() {
		return _verkehrsAnalyse;
	}

	RDSParameterDaten getParameter() {
		return getVerkehrsanalyse().getParameter();
	}

	ClientDavInterface getConnection() {
		return getVerkehrsanalyse().getConnection();
	}

	/**
	 * Konstruiert eine Instanz vom Typ <code>StoerfallSituation</code>.
	 * 
	 * @param segment
	 *            das korrespondierende Strassensegment
	 * @param verkehrsAnalyse
	 *            {@link Verkehrsanalyse}
	 */
	public StoerfallSituation(final StrassenSegment segment, final Verkehrsanalyse verkehrsAnalyse) {
		_segment = segment;
		_verkehrsAnalyse = verkehrsAnalyse;
	}

	/**
	 * Fügt dieser <code>StoerfallSituation</code> einen <code>Stoerfall</code>
	 * hinzu.
	 * 
	 * @param stoerfall
	 *            der hinzuzufügende <code>Stoerfall</code>
	 */
	void addStoerfall(final Stoerfall stoerfall) {
		_stoerfaelle.add(stoerfall);
	}

	/**
	 * Entfernt einen <code>Stoerfall</code> aus dieser
	 * <code>StoerfallSituation</code>.
	 * 
	 * @param stoerfall
	 *            der zu entfernende <code>Stoerfall</code>
	 */
	void removeStoerfall(final Stoerfall stoerfall) {
		_stoerfaelle.remove(stoerfall);
	}

	/*
	 * (Kein Javadoc)
	 * 
	 * @see de.bsvrz.ste.rdstmc.verkehrsanalyse.StoerfallListener
	 * #neuerStoerfall(de.bsvrz.ste.rdstmc.verkehrsanalyse.Stoerfall,
	 * de.bsvrz.ste.rdstmc.verkehrsanalyse.StoerfallDaten)
	 */
	public void neuerStoerfall(final Stoerfall quelle, final StoerfallDaten daten) {
		if (_stoerfaelle.contains(quelle)) {
			synchronized (getVerkehrsanalyse().mutexStoerfallSituation) {
				final StoerfallDaten aktuell = _stoerfallDaten.get(quelle);

				if (daten != null) {
					if (aktuell == null || !daten.equals(aktuell)) {
						_stoerfallDaten.put(quelle, daten);
						_debug.finer("Vor-Überlagerung im Segment " + getSegment() + " bei " + quelle.getObjekt() + ":"
								+ "\nAlter Zustand: "
								+ (aktuell == null ? "<KEINER> "
										: (" \"" + aktuell.zustand + "\" - Priorität " + aktuell.prioritaet + " von "
												+ aktuell.start + "m bis " + aktuell.ende + "m"))
								+ "\nNeuer Zustand: \"" + daten.zustand + "\" - Priorität " + daten.prioritaet + " von "
								+ daten.start + "m bis " + daten.ende + "m");
						/* Stau-Plausibilität: ALT Anfang */
						ueberlagern();
						/* Stau-Plausibilität: ALT Ende */
						fireStoerfallSituationListener();
					} else {
						_debug.finer("SIZE: " + _ueberlagerteDaten.size());
						_debug.finer("Vor-Überlagerung unverändert im Segment " + getSegment() + " bei "
								+ quelle.getObjekt() + ":" + "\nZustand \"" + daten.zustand + "\"");
					}
				} else {
					if (aktuell != null) {
						_stoerfallDaten.remove(quelle);

						_debug.finer("Vor-Überlagerung im Segment " + getSegment() + " bei " + quelle.getObjekt() + ":"
								+ "\nAlter Zustand: \"" + aktuell.zustand + "\" - Priorität " + aktuell.prioritaet
								+ " von " + aktuell.start + "m bis " + aktuell.ende + "m"
								+ "\nNeuer Zustand: <KEINER>");
						/* Stau-Plausibilität: ALT Anfang */
						ueberlagern();
						/* Stau-Plausibilität: ALT Ende */
						fireStoerfallSituationListener();
					}
				}
			}
		} else {
			throw new IllegalArgumentException("Störfall für " + quelle.getObjekt()
					+ " ist nicht in Störfallsituation für Segment " + _segment + " enthalten");
		}
	}

	private void ueberlagern_neu(final StoerfallDaten daten) {

		// Nur Daten mit positiver Länge werden berücksichtigt
		if (!(daten.start < daten.ende)) {
			return;
		}

		// Entferne alle Elemente, die von dem Element überlappt werden
		int indexGlobal = -1;
		int indexRelevant = -1;
		StoerfallDaten erstes = null;
		StoerfallDaten letztes = null;
		final ListIterator<StoerfallDaten> it = _ueberlagerteDaten.listIterator();

		while (it.hasNext()) {
			final StoerfallDaten lauf = it.next();
			if (lauf.start >= daten.ende) {
				break;
			}
			indexGlobal += 1;
			if ((daten.start >= lauf.start && daten.start < lauf.ende)
					|| (daten.ende > lauf.start && daten.ende <= lauf.ende)
					|| (daten.start < lauf.start && daten.ende > lauf.ende)) {
				if (erstes == null) {
					erstes = lauf; // erstes überlapptes Element
					indexRelevant = indexGlobal;
				}
				letztes = lauf; // jeweils zuletzt überlapptes Element
				// index += 1; // und dessen Index
				it.remove();
			}
		}

		// Falls das letzte Element der überlappten Elemente hinten über den
		// Stau hinausragt,
		// füge ein entsprechend passendes Element nach dem Stau ein
		// if (letztes != null && letztes.ende > daten.ende)
		// {
		// daten.add(indexRelevant, new StoerfallDaten(letztes.zustand,
		// stau.endOffset, letztes.ende, letztes.lokal));
		// }
		// // Füge eigentliches Stau-Element ein
		// daten.add(indexRelevant >= 0 ? indexRelevant : indexGlobal + 1,
		// new StoerfallSituationDaten(StoerfallZustand.STAU, stau.startOffset,
		// stau.endOffset, null));
		// // Falls das erste Element der überlappten Elemente vorne über den
		// Stau hinausragt,
		// // füge ein entsprechend passendes Element vor dem Stau ein
		// if (erstes != null && erstes.start < stau.startOffset)
		// {
		// daten.add(indexRelevant, new StoerfallSituationDaten(erstes.zustand,
		// erstes.start, stau.startOffset, erstes.lokal));
		// }
		// }
		// String text = "Störfallsituation " + getSegment() + " nach
		// Stauüberlagerung: ";
		// for (StoerfallSituationDaten lauf: daten)
		// {
		// text += "\n" + lauf.zustand + " von " + lauf.start + "m bis " +
		// lauf.ende +"m";
		// }
		// _debug.fine(text);
		// }

	}

	private void ueberlagern(final StoerfallDaten daten) {
		// Nur Daten mit positiver Länge werden berücksichtigt
		// if (!(daten.start < daten.ende))
		// Änderung LST: Start darf nur nicht groesser als Ende sein (MQ's haben
		// keine Ausdehnung).
		if (daten.start > daten.ende) {
			return;
		}

		final LinkedList<StoerfallDaten> relevanteDatenIst = new LinkedList<StoerfallDaten>();
		int indexGlobal = -1;
		int indexRelevant = -1;

		// Bilde die Liste relevanteDatenIst, die alle überlappenden Elemente
		// der bisherigen Überlagerung enthält
		for (StoerfallDaten lauf : _ueberlagerteDaten) {
			if (lauf.start >= daten.ende) {
				break;
			}
			indexGlobal += 1;
			if ((daten.start >= lauf.start && daten.start < lauf.ende)
					|| (daten.ende > lauf.start && daten.ende <= lauf.ende)
					|| (daten.start < lauf.start && daten.ende > lauf.ende)) {

				if (indexRelevant == -1) {
					indexRelevant = indexGlobal;
				}
				relevanteDatenIst.add(lauf);
			}
		}

		if (!relevanteDatenIst.isEmpty()) {
			StoerfallDaten soll = daten;
			final List<StoerfallDaten> relevanteDatenSoll = new LinkedList<StoerfallDaten>();
			final Iterator<StoerfallDaten> it = relevanteDatenIst.iterator();

			while (soll.ende >= soll.start && it.hasNext()) {
				StoerfallDaten ist = it.next();

				if (ist.start < daten.start) {
					// Die vorhanden Daten ragen über die einzufügenden Daten am
					// Anfang heraus
					relevanteDatenSoll.add(
							new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet, ist.start, daten.start));
					ist = new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet, daten.start, ist.ende);
				} else if (soll.start < ist.start) {
					// Die einzufügenden Daten ragen über die vorhanden Daten am
					// Anfang heraus
					relevanteDatenSoll.add(
							new StoerfallDaten(soll.stoerfall, soll.zustand, soll.prioritaet, soll.start, ist.start));
					soll = new StoerfallDaten(soll.stoerfall, soll.zustand, soll.prioritaet, ist.start, soll.ende);
				}

				if (soll.prioritaet > ist.prioritaet
						|| (soll.prioritaet == ist.prioritaet && soll.zustand.compareTo(ist.zustand) > 0)) {
					if (soll.ende > ist.ende) {
						relevanteDatenSoll.add(new StoerfallDaten(soll.stoerfall, soll.zustand, soll.prioritaet,
								soll.start, ist.ende));
						soll = new StoerfallDaten(soll.stoerfall, soll.zustand, soll.prioritaet, ist.ende, soll.ende);
						if (!it.hasNext()) {
							relevanteDatenSoll.add(soll);
						}
					} else {
						relevanteDatenSoll.add(soll);
						// Evtl. Überstand wird behandelt
						if (ist.ende > soll.ende) {
							relevanteDatenSoll.add(new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet,
									soll.ende, ist.ende));
						}
					}
				} else {
					if (ist.ende >= soll.ende) {
						relevanteDatenSoll.add(
								new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet, ist.start, soll.ende));
						// Evtl. Überstand wird behandelt
						if (ist.ende > soll.ende) {
							relevanteDatenSoll.add(new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet,
									soll.ende, ist.ende));
						}
					} else {
						relevanteDatenSoll.add(
								new StoerfallDaten(ist.stoerfall, ist.zustand, ist.prioritaet, ist.start, ist.ende));
						soll = new StoerfallDaten(soll.stoerfall, soll.zustand, soll.prioritaet, ist.ende, soll.ende);
						if (!it.hasNext()) {
							relevanteDatenSoll.add(soll);
						}
					}
				}
			}

			indexRelevant = indexRelevant >= 0 ? indexRelevant : indexGlobal + 1;
			_ueberlagerteDaten.removeAll(relevanteDatenIst);

			StoerfallDaten element = null;
			final ListIterator<StoerfallDaten> iter = relevanteDatenSoll.listIterator();
			while (iter.hasNext()) {
				final StoerfallDaten lauf = iter.next();
				if (element == null) {
					element = new StoerfallDaten(lauf.stoerfall, lauf.zustand, lauf.prioritaet, lauf.start, lauf.ende);
				} else {
					if (element.zustand.compareTo(lauf.zustand) == 0
							&& element.stoerfall.getDavObjekt().getId() == lauf.stoerfall.getDavObjekt().getId()
							&& element.prioritaet == lauf.prioritaet) {
						element = new StoerfallDaten(element.stoerfall, element.zustand, element.prioritaet,
								element.start, lauf.ende);
					} else {
						_ueberlagerteDaten.add(indexRelevant, element);
						indexRelevant += 1;
						element = new StoerfallDaten(lauf.stoerfall, lauf.zustand, lauf.prioritaet, lauf.start,
								lauf.ende);
					}
				}
			}
			if (element != null) {
				_ueberlagerteDaten.add(indexRelevant, element);
			}

			// _ueberlagerteDaten.addAll(indexRelevant, relevanteDatenSoll);

		} else {
			_ueberlagerteDaten.add(indexGlobal + 1, daten);
		}
	}

	public Collection<Stoerfall> getStoerFaelle() {
		return Collections.unmodifiableCollection(_stoerfaelle);
	}

	public void addStoerfallSituationListener(final StoerfallSituationListener listener) {
		synchronized (_stoerfallSituationListener) {
			_stoerfallSituationListener.add(listener);
		}
	}

	public void renmoveStoerfallSituationListener(final StoerfallSituationListener listener) {
		synchronized (_stoerfallSituationListener) {
			_stoerfallSituationListener.remove(listener);
		}
	}

	/**
	 * Benachrichtige alle angemeldeten Listener vom Typ
	 * <code>StoerfallSituationDaten</code>
	 */
	public void fireStoerfallSituationListener() {
		/* Stau-Plausibilität: NEU Anfang */
		// ueberlagern();
		/* Stau-Plausibilität: NEU Ende */
		List<StoerfallSituationDaten> ueberlagerteDaten = ueberlagereStaus();
		final List<StoerfallSituationDaten> daten = bestimmeStoerfallSituationDaten(ueberlagerteDaten);

		synchronized (_stoerfallSituationListener) {
			for (StoerfallSituationListener listener : _stoerfallSituationListener) {
				listener.neueStoerfallSituation(this, daten);
			}
		}
	}

	private List<StoerfallSituationDaten> ueberlagereStaus() {
		final List<StoerfallSituationDaten> daten = new LinkedList<StoerfallSituationDaten>();

		for (StoerfallDaten lauf : _ueberlagerteDaten) {
			daten.add(new StoerfallSituationDaten(lauf.zustand, lauf.start, lauf.ende,
					lauf.stoerfall instanceof LokalerStoerfall ? lauf.stoerfall : null));
		}

		if (getParameter().getRDSBeruecksichtigeStauverlaufsanalyse()) {
			for (SegmentSituationDaten stau : getVerkehrsanalyse().getStauDaten(getSegment())) {
				// Nur Stauanteile positiver Länge werden berücksichtigt
				if (!(stau.startOffset < stau.endOffset)) {
					continue;
				}
				// Entferne alle Elemente, die von dem Stau überlappt werden
				int indexGlobal = -1;
				int indexRelevant = -1;
				StoerfallSituationDaten erstes = null;
				StoerfallSituationDaten letztes = null;
				final ListIterator<StoerfallSituationDaten> it = daten.listIterator();

				while (it.hasNext()) {
					final StoerfallSituationDaten lauf = it.next();
					if (lauf.start >= stau.endOffset) {
						break;
					}
					indexGlobal += 1;
					if ((stau.startOffset >= lauf.start && stau.startOffset < lauf.ende)
							|| (stau.endOffset > lauf.start && stau.endOffset <= lauf.ende)
							|| (stau.startOffset < lauf.start && stau.endOffset > lauf.ende)) {
						if (erstes == null) {
							erstes = lauf; // erstes überlapptes Element
							indexRelevant = indexGlobal;
						}
						letztes = lauf; // jeweils zuletzt überlapptes Element
						// index += 1; // und dessen Index
						it.remove();
					}
				}

				// Falls das letzte Element der überlappten Elemente hinten über
				// den Stau hinausragt,
				// füge ein entsprechend passendes Element nach dem Stau ein
				if (letztes != null && letztes.ende > stau.endOffset) {
					daten.add(indexRelevant,
							new StoerfallSituationDaten(letztes.zustand, stau.endOffset, letztes.ende, letztes.lokal));
				}
				// Füge eigentliches Stau-Element ein
				daten.add(indexRelevant >= 0 ? indexRelevant : indexGlobal + 1,
						new StoerfallSituationDaten(StoerfallZustand.STAU, stau.startOffset, stau.endOffset, null));
				// Falls das erste Element der überlappten Elemente vorne über
				// den Stau hinausragt,
				// füge ein entsprechend passendes Element vor dem Stau ein
				if (erstes != null && erstes.start < stau.startOffset) {
					daten.add(indexRelevant,
							new StoerfallSituationDaten(erstes.zustand, erstes.start, stau.startOffset, erstes.lokal));
				}
			}
			String text = "Störfallsituation " + getSegment() + " nach Stauüberlagerung: ";
			for (StoerfallSituationDaten lauf : daten) {
				text += "\n" + lauf.zustand + " von " + lauf.start + "m bis " + lauf.ende + "m";
			}
			_debug.fine(text);
		}

		return daten;
	}

	/**
	 * Bestimmt aus den überlagerten Daten die Folge der
	 * StoerfallSituationsDaten mit dem schärfsten Zustand.
	 * 
	 * @param ueberlagerteDaten
	 *            Liste von {@link StoerfallSituationDaten}
	 * 
	 * @return Liste der StoerfallSituationsDaten mit dem schärfsten Zustand
	 */
	private List<StoerfallSituationDaten> bestimmeStoerfallSituationDaten(
			final List<StoerfallSituationDaten> ueberlagerteDaten) {
		final List<StoerfallSituationDaten> result = new LinkedList<StoerfallSituationDaten>();

		StoerfallSituationDaten daten = null;
		for (StoerfallSituationDaten lauf : ueberlagerteDaten) {
			if (daten == null) {
				daten = new StoerfallSituationDaten(lauf.zustand, lauf.start, lauf.ende, lauf.lokal);
			} else {
				if (daten.zustand.compareTo(lauf.zustand) == 0 && daten.ende >= lauf.start) {
					daten = new StoerfallSituationDaten(lauf.zustand, daten.start, lauf.ende,
							lauf.lokal == daten.lokal ? lauf.lokal : null);
				} else {
					result.add(daten);
					daten = new StoerfallSituationDaten(lauf.zustand, lauf.start, lauf.ende, lauf.lokal);
				}
			}
		}
		if (daten != null) {
			result.add(daten);
		}

		return result;
	}

	private void ueberlagern() {
		_ueberlagerteDaten.clear();

		for (StoerfallDaten daten : _stoerfallDaten.values()) {
			/* Stau-Plausibilität: NEU Anfang */
			// if (daten.zustand.compareTo(StoerfallZustand.STAU) == 0 &&
			// getParameter().getRDSPlausibStauverlaufsanalyse())
			// {
			// if (!getVerkehrsanalyse().existiertStau(getSegment(),
			// daten.start, getSegment(), daten.ende))
			// {
			// // Nichtplausible Stau-SI-Zustände werden beim Überlagern nicht
			// berücksichtigt
			// continue;
			// }
			// }
			/* Stau-Plausibilität: NEU Ende */

			ueberlagern(daten);
		}

		/*
		 * 20090402: Der folgende Abschnitt verursacht OutOfMemoryError und wird
		 * nicht mehr verwendet
		 * 
		 * String text = "Zustände nach Überlagerung im Segment " + _segment +
		 * ":"; if (!_ueberlagerteDaten.isEmpty()) { for (StoerfallDaten daten:
		 * _ueberlagerteDaten) { text += "\n\"" + daten.zustand + "\" - " +
		 * daten.start + "m bis " + daten.ende + "m von " +
		 * daten.stoerfall.getObjekt() + " mit Priorität " + daten.prioritaet; }
		 * } else { text += " <KEINE>"; } _debug.finer(text);
		 */
	}

}
